# This file contains the delegator functions to the actual data layer code.

# Function takes in a relational table name and returns the attribute names.
#
# Args: 
#  dataset.name : Name of the relational table.
#
# Returns:
#  List of attribute names.
GetAttrnames <- function(dataset.name) {
	# Error handling.
	if(is.null(dataset.name)) {
		stop('Argument dataset.name cannot be null')
	}
	
	res <- bGetAttrnames(dataset.name)
	return(res)
}


# Function takes in session information and updates the table in the db
#
# Args:
# session.id : Current session identifier
# user.id : Current user id
#
# Returns:
# Response from db

UpdateSessionInfo <- function(session.id, user.id) {
	
	# Error handling.
	if(is.null(session.id) || is.null(user.id)) {
		stop('Argument session.id or user.id cannot be null')
	}
	
	res <- bUpdateSessionInfo(session.id, user.id)
	return(res)
}


# Function that stores program text in to the oracle db.
#
# Args:
# session.id : Current session identifier
# program.id : Program identifier << global >>
# program.name : Program name
# program.file : Filename where the program is stored.
#
# Returns
# Response from dB

UpdateProgramInfo <- function(session.id,  program.id, program.name, file.name, dataset.id = 1) {
	
	# Error handling
	if(is.null(session.id) || is.null(program.id)) {
		stop('Argument session.id or program.id cannot be null')
	}
	
	if(is.null(program.name) || is.null(file.name)) {
		stop('Argument program.name or file.name cannot be null')
	}
	

	lines.from.file <- readLines(file.name, encoding = 'UTF-8')
	old.pattern <- paste0('dataset.id = ', dataset.id)
	new.pattern <- paste0('dataset.id = !!')
	lines.from.file <- sub(old.pattern, new.pattern, lines.from.file)
	str.program <- paste(lines.from.file, collapse="@@")
	
	
	res <- bUpdateProgramInfo(session.id, program.id, program.name, str.program)
	
	return(res)

}


UpdateProgramDataset <- function(program.id, dataset.id) {

	# Error handling
	if(is.null(program.id) || is.null(dataset.id)) {
		stop('Argument program.id or dataset.id cannot be null')
	}
	
	res <- bUpdateProgramDataset(program.id, dataset.id)
	return(res)
	
}

# Function to retrieve the program text given the program name
#
# Args:
# program.name : Name of the program
#
# Returns:
# program.text : Text of the program.

GetPgmTextFromName <- function(program.name) {

	# Error handling.
	if(is.null(program.name)) {
		stop('Argument program.name cannot be null');
	}
	
	res <- bGetPgmTextFromName((program.name))
	res <- strsplit(res, "@@")[[1]]
	return(res)

}


# Function to update the table names in the mdb info, and store the mdb tables
#
# Args:
# program.id : Program identifier
# session.id : Session identifier
# program.name : Name of the program
#
# Returns:
# Reponse from the database
UpdateMdbTableNames <- function(program.id, session.id, program.name, mdb, dataset.id) {
	
	# Error handling.
	if(is.null(program.id) || is.null(session.id) || is.null(program.name)) {
		stop('Argument program.id, session.id and program.name cannot be null')
	}

	if(is.null(mdb)) {
		stop('Argument mdb cannot be null')
	}

	res <- StoreMdbTables(program.id, session.id, tolower(program.name), mdb, dataset.id)
	#res <- bUpdateMdbTableNames(program.id, session.id, tolower(program.name))
	return(res)
}

# Function to remove the table names from the mdbinfo, and remove the mdb tables from oracle.
#
# Args:
# program.id : Program identifier
# session.id : Session identifier
# program.name : Name of the program
#
# Returns:
# Response from the database

RemoveMdbTableNames <- function(program.id, session.id, program.name) {
	
	# Error handling.
	if(is.null(program.id) || is.null(session.id) || is.null(program.name)) {
		stop('Argument program.id, session.id and program.name cannot be null')
	}

	program.name <- tolower(program.name)
	res <- RemoveMdbTables(program.id, session.id, program.name)
	res <- bDeleteTableNames(program.id)

	return(res)
}

WriteTable <- function(conn, name, value, ...) {
	print(name)
	if(nrow(value) > 0) {
		if(dbExistsTable(conn, name) == TRUE)
			dbRemoveTable(conn, name)
		dbWriteTable(conn, name, value, ...)
	}
}


# Function to store the mdb tables in the oracle db
#
# Args: 
# program.id : Program identifier
# session.id : Session identifier
# program.name : Name of the program
# mdb : Data has that contains the data.tables to be stored.
#
# Returns:
# None
StoreMdbTables <- function(program.id, session.id, program.name, mdb, dataset.id) {
	res <- GetMDBTableNames(program.id, session.id, program.name, dataset.id)
	
	WriteTable(mconn, res$best.k.config, mdb$BestKConfig)
	WriteTable(mconn, res$cg.config, mdb$CgConfig)
	WriteTable(mconn, res$dataset, mdb$Dataset)
	WriteTable(mconn, res$derived.datasets, mdb$DerivedDatasets)
	WriteTable(mconn, res$derived.features, mdb$DerivedFeatures)
	WriteTable(mconn, res$featureset_db, mdb$FeatureSetDB)
	WriteTable(mconn, res$features, mdb$Features)
	WriteTable(mconn, res$igd.config, mdb$IgdConfig)
	WriteTable(mconn, res$program, mdb$Program)
	for(i in 1:length(res$featuremap_db)) {
		WriteTable(mconn, res$featuremap_db[[i]], mdb$FeatureMapDB[i])
	}
}

# Function to remove the mdb tables from oracle db
#
# Args: 
# program.id : Program identifier
# session.id : Session identifier
# program.name : Name of the program
#
# Returns:
# None
RemoveMdbTables <- function(program.id, session.id, program.name, dataset.id) {
	res <- GetMDBTableNames(program.id, session.id, program.name, dataset.id)
	
	RemoveTable(mconn, res$best.k.config)
	RemoveTable(mconn, res$cg.config)
	RemoveTable(mconn, res$dataset)
	RemoveTable(mconn, res$derived.datasets)
	RemoveTable(mconn, res$derived.features)
	RemoveTable(mconn, res$featuremap.)
	RemoveTable(mconn, res$featureset_)
	RemoveTable(mconn, res$features)
	RemoveTable(mconn, res$igd.config)
	RemoveTable(mconn, res$program)
}

RemoveTable <- function(conn, table.name) {
	
	if(is.null(conn) || is.null(table.name)) {
		return(NULL)
	}
	if(dbExistsTable(conn, table.name)) {
		dbRemoveTable(conn, table.name)
	} else {
		printf("Table %s does not exist", table.name)
	}
}

SaveSession <- function(program.name, dataset.id, update = TRUE) {
	SetProgramname(program.name)
	program.id <- GetMaxVals('PROGRAM_INFO', 'PROGRAM_ID') + 1
	if(is.na(program.id))
		program.id <- 1
	SetProgramId(program.id)
	StoreAll(mdb, dataset.id, update)
}

StoreAll<- function(mdb, dataset.id, update = TRUE) {
	program.id <- GetProgramId()
	session.id <- GetSessionId()
	prog.name <- GetProgramname()
	print("Updation done")
	str <- paste(unlist(mdb$program.lines))
	writeLines(str, con = 'tttemp')
	file.name <- 'tttemp'
	StoreMDB(prog.name, dataset.id, mdb)
	if(update == TRUE) {
		UpdateProgramInfo (session.id,  program.id, prog.name, file.name, dataset.id) 
		#UpdateMdbTableNames(program.id, session.id, prog.name, mdb, dataset.id)
		UpdateProgramDataset(program.id, dataset.id)
		#StoreProvResults()
	}
}

GetRepos <- function() {
	repos <- '/home/oracle/r/demo/repos'
}

StoreMDB <- function(prog.name, dataset.id, mdb) {
	repos <- GetRepos()
	file.name <- paste0(repos, "/", prog.name, "_", dataset.id, ".RData")
	save(file = file.name, "mdb")
	
}
	
GetMaxVals <- function(dataset.name, attr.vec) {
	bMaxValues(dataset.name, attr.vec)
}

GetMinVals <- function(dataset.name, attr.vec) {
    	bMinValues(dataset.name, attr.vec)
}
UpdateFeaturesTable <- function(dataset.name) {
	attr.vec <- GetAttrnames(dataset.name)
	max.val <- GetMaxVals(dataset.name, attr.vec)
	max.val <- GetMinVals(dataset.name, attr.vec)
}

GetMaxMinVals <- function(dataset.name, attr.vec) {
	max.val <- GetMaxVals(dataset.name, attr.vec)
	min.val <- GetMinVals(dataset.name, attr.vec)
	return(list(max.val = max.val, min.val = min.val))
}

GetCompatiblePrograms <- function(dataset.id) {
	bGetCompatiblePrograms(dataset.id)
}

GetDatasetsFromDB <- function() {
	bGetDatasets()
}

ExecuteQuery <- function(query) {
	bExecuteQuery(query)
}

GetTablesInDateRange <- function(dataset.name, st_date, en_date) {
	if(is.na(st_date) || is.na(en_date))
		return()
	bGetTablesInDateRange(dataset.name, st_date, en_date)
}

UpdateNumTimesAccessed <- function(dataset.name) {
	datasets <- bGetDatasets()
	datasets <- datasets[, "TABLE_NAME"]
	
	if(dataset.name %in% datasets) {
		num.times <- bGetNumTimesAccessed(dataset.name)
		num.times <- num.times + 1
		bUpdateNumTimesAccessed(dataset.name, num.times)
	
	}
	else {
		#derived dataset
	}
		
}
