# Adding a row to a data.table
# 
# Args:
#  x : Existing data table
#  r : Row to be inserted to the existing data table.
#
# Returns:
#  x : Updated data.table

AddRow <- function(x, r) {

	# Set the column names
	if(length(colnames(x)) > 0) {
		setnames(r, colnames(r), colnames(x))
	}
	
	# Do a row bind.
	x <- rbind(x, r)
	return(x)
}

# Get column names for a table
#
# Args:
#  table.name : String, name of a table.
# 
# Returns:
#  col.names  : Column names for the input table name.

GetColumnNames <- function(table.name) {
	
	# Error handling
	if (length(table.name) == 0) {
		stop('Argument table.name cannot be of length 0')
	}
	
	# Convert the table.name to lower case, to match the strings ignoring case
	table.name <- tolower(table.name)
	
	# Do a switch and return appropriate column names
	switch(table.name, 
		program = c('dataset.id', 'op.id', 'step.num', 'program.ctr', 'inpset.id1', 'inpset.id2', 'inplist.id1', 'outset.id', 'outlist.id'),
		featuresetdb = c('inst.id', 'set.val'),
		featuremapdb = c('inst.id', 'feature.list', 'values'),
		#cgconfig = c('dataset.id', 'step.num', 'program.ctr', 'num.iters', 'init.grad', 'init.wt','init.dir', 'num.folds'),
		cgconfig = c('dataset.id', 'step.num', 'program.ctr', 'num.iters', 'init.wt', 'num.folds'),
		igdconfig = c('dataset.id', 'step.num', 'program.ctr', 'num.iters', 'step.size',
'decay', 'init.wt', 'num.folds'),
		bestkconfig = c('dataset.id', 'step.num', 'program.ctr', 'k.val'),
		derivedfeatures = c('dataset.id', 'step.num', 'program.ctr', 'feature.id1', 'feature.id2', 'newfeature.id'),
		dataset = c('name', 'initial.num.features', 'target.index'),
		deriveddatasets = c('dataset.id', 'feature.indices', 'table.name'),
		features = c('feature.name', 'feature.id', 'dataset.id', 'min.val', 'max.val'),
		operators = c('op.name', 'op.id', 'table.name'),
		compatibleprograms = c('dataset.id', 'program.name', 'file.name')
	)
		

}

# Inserting a Row to a table
#
# Args:
#  table.name : Name of the table in the MDB data structure.
#   ...       : Args corresponding to the table.   
# Returns:
#   Number of rows in the updated table

InsertRowInTable <- function(table.name, ...) {

	# Error handling
	if (TRUE == is.null(table.name)) {
		stop('Argument table.name cannot be NULL')
	}
	
	mdb[[table.name]] <- AddRow(mdb[[table.name]], ...)
	if(nrow(mdb[[table.name]]) == 1) {
		# Get the column names
		col.names <- GetColumnNames(table.name)
		
		# Set the column names
		setnames(mdb[[table.name]], colnames(mdb[[table.name]]), col.names)
	}
	return(nrow(mdb[[table.name]]))
}

# Given dataset name, return its dataset id.
#
# Args:
#  dataset.name : dataset name
# Returns:
#  Positional occurence of dataset.name in the dataset table stored in MDB.

GetDatasetId <- function(dataset.name) {
	
	# Error handling
	if (is.null(dataset.name)) {
		stop('Argument dataset.name cannot be NULL')
	}
	
	idx <- which(mdb[['Dataset']]$name == dataset.name)

	if (length(idx) == 0) {
	
		out.str <- sprintf('dataset.name ( %s ) not found', dataset.name)
		warning(out.str)
		ret.id <- NULL

	} else {
		
		ret.id <- idx
	}
	
	return(ret.id)
}

RegisterAllDatasets <- function() {
	res <- GetDatasetsFromDB()
	for(i in 1:nrow(res)) {
		RegisterDataset(res[i, "TABLE_NAME"], res[i, "INIT_NUMFEATURES"], res[i,
"TARGET_FEATURE_INDEX"])
	} 
	
}

# Registering a dataset.
# Args:
#  dataset.name : Table name in the db.
#  num.features : Total number of features (including the target).
#  target.index : Target index. 

RegisterDataset <- function(dataset.name, num.features, target.index) {
	
	# Error handling
	if(is.null(dataset.name)) {
		stop('Argument dataset.name cannot be NULL')
	}
	if(num.features <= 0) {
		stop('Argument num.features should be positive')
	}
	if(target.index < 0 ) {
		stop('Argument target.index should be positive ')
		
	}
	
	# Get the attribute names for the relational table
	attr.vec <- GetAttrnames(dataset.name)

	max.val <- GetMaxVals(dataset.name, attr.vec)
	min.val <- GetMinVals(dataset.name, attr.vec)

	
	
	# Insert dataset.name in to dataset table.
	
	# TODO(pradap) : Check if the dataset is already inserted.
	InsertRowInTable('Dataset', data.table(dataset.name, num.features, target.index))
	
	# Get the Dataset id.
	dataset.id <- GetDatasetId(dataset.name)

	if(dataset.id == MNULL) {
		stop('dataset.id is -1')
	}	
	
	# Insert in to the Derived datasets table
	feat.set <- setdiff(c(0:(num.features - 1)), target.index)
	InsertRowInTable('DerivedDatasets', data.table(dataset.id, list(feat.set), dataset.name))

	# Update in the feature table
	for(idx in 1:num.features) {
		InsertRowInTable('Features', data.table(attr.vec[idx], as.character(idx - 1),
as.character(dataset.id), as.numeric(unlist(min.val[idx])), as.numeric(unlist(max.val[idx]))))
		#InsertRowInTable('Features', data.table(attr.vec[idx], as.numeric(idx - 1),
#as.numeric(dataset.id), as.numeric(unlist(min.val[idx])), as.numeric(unlist(max.val[idx]))))
	}

	# Update in the Compatible Programs Table
	# For a dataset id get the compatible programs from the database
	#program.names <- GetCompatiblePrograms(dataset.id)
	return(dataset.id)
	
	
	
}
# Given feature names, returns the feature ids from the dataset.
# 
# Args:
#  feature.names : Name of the features from the dataset.
#  dataset.id : Dataset id.
#
# Returns:
#  Feature Ids corresponding to the feature names from the dataset. 

GetFeatureIds <- function(feature.names, dataset.id) {
	
	# Get the valid indices from the MDB table.
	valid.indices <- which(mdb[['Features']]$dataset.id == dataset.id)
	
	# Initialize the output vector.
	feat.ids <- c()
	
	# Ids are indexed from 0.
	for(i in 1:length(feature.names)) {
		id <- which(mdb$Features[valid.indices]$feature.name == feature.names[i])
		if(length(id) == 0) {
			stop('Feature name did not match a feature.id')
		}
		sel.feature.id <- mdb$Features[id]$feature.id
		feat.ids <- union(feat.ids, sel.feature.id)
	}
	return(feat.ids)
	
}

# Get the current step number from the program table
#
# Args:
#  NONE
# 
# Returns:
#  Current step number from the program table

GetStepNum <- function() {
	return(nrow(mdb$Program))
}

GetInputSetId.1 <- function(step.num) {
	return(mdb$Program[step.num]$inpset.id1)
}

GetInputSetId.2 <- function(step.num) {
	return(mdb$Program[step.num]$inpset.id2)
}

GetInputList.1 <- function(step.num) {
	return(mdb$Program[step.num]$inplist.id1)
}

GetOutputSetId <- function(step.num) {
	return(mdb$Program[step.num]$outset.id)
}

GetOutputListId <- function(step.num) {
	return(mdb$Program[step.num]$outlist.id)
}

GetFeatureSet <- function(feat.setid) {
	return(mdb$FeatureSetDB[feat.setid]$set.val)
}

GetFeatureMapValues <- function(list.id) {
	return(mdb$FeatureMapDB[list.id]$values)
}
GetFeatureMapFeatureSet <- function(list.id) {
	return(mdb$FeatureMapDB[list.id]$feature.list)
}
GetFeatureMap <- function(list.id) {
	return(mdb$FeatureMapDB[list.id])
}

SetFeatureSet <- function(feat.setid, set.val) {
	mdb$FeatureSetDB[feat.setid]$set.val <- list(list(set.val))
}

SetFeatureMapValues <- function(list.id, list.val) {
	mdb$FeatureMapDB[list.id]$values <- list(list(list.val))
}
SetFeatMapFeatSetList <- function(list.id, featset.list) {
	mdb$FeatureMapDB[list.id]$feature.list <- list(list(featset.list))
}

CopyFeatureSet <- function(set.id.from, set.id.to) {
	if((set.id.from == MNULL) || (set.id.to == MNULL))
		stop('Arguments set.id.from/to cannot be -1')
	set.val <- GetFeatureSet(set.id.from)
	mdb$FeatureSetDB[set.id.to]$set.val <- set.val
}

CopyFeatureMapFeatSet <- function(list.id.from, list.id.to) {
	if((list.id.from == MNULL) || (list.id.to == MNULL))
		stop('Arguments set.id.from/to cannot be -1')
	feat.set <- GetFeatureMapFeatureSet(list.id.from)
	mdb$FeatureMapDB[list.id.to]$feature.list <- list(feat.set)
	
}

CopyFeatureMapValues <- function(list.id.from, list.id.to) {
	if((list.id.from == MNULL) || (list.id.to == MNULL))
		stop('Arguments set.id.from/to cannot be -1')
	
	values <- GetFeatureMapValues(list.id.from)
	mdb$FeatureMapDB[list.id.to]$values <- list(values)
	
}

is.interactive <- function() {
	return(mdb$interactive)
	
}

SetInteractive <- function(bool.in) {
  mdb$interactive <- bool.in
}

GetNewFeatureName <- function(dataset.id) {
		
	v <- which(mdb$Features$dataset.id == dataset.id)
	m <- max(as.integer(mdb$Features$feature.id[v])) + 2
	feature.name <- paste('F', m, sep = "_")
	feat.id <- m - 1
	return(list(feature.name = feature.name, feat.id = feat.id))
}


GetFeatureNames <- function(feature.ids, dataset.id) {
	
	valid.indices <- which(mdb$Features$dataset.id == dataset.id)
	feature.names <- c()

	# Error handling
	if(all(intersect(feature.ids, mdb$Features[valid.indices]$feature.id) == feature.ids) ==
FALSE) {
		stop("Argument feature.ids do not match with feature indices from the dataset")
	}	
	
	for(idx in 1:length(feature.ids)) {
		idx <- which(mdb$Features[valid.indices]$feature.id == feature.ids[idx])
		feature.names <- c(feature.names, mdb$Features[valid.indices[idx]]$feature.name)
	}
	
	return(feature.names)
}

AddFeatureToDataset <- function(feat.name, feat.id, dataset.id) {
	
	InsertRowInTable('Features', data.table(feat.name, feat.id, dataset.id))
}

is.data.aware <- function(step.num) {
	
	op <- tolower(mdb$Program[step.num, ]$op.id)
	switch(op, 
		feature_set = FALSE,
		feature_map = FALSE,
		add_feature = FALSE,
		del_feature = FALSE,
		intersect_feature = FALSE,
		best_k = FALSE,
		create_feature = FALSE,
		correlation_x = TRUE,
		correlation_y = TRUE,
		coef_igd_train = TRUE,
		coef_cg_train = TRUE,
		evaluate = TRUE,
		step_add = TRUE,
		step_del = TRUE,
		eval_cg_train = TRUE,
		eval_igd_train = TRUE,
		eval_cv_test = TRUE,
		eval_aic_test = TRUE,
		stepadd_eval_cg_train = TRUE,
		stepadd_eval_igd_train = TRUE,
		stepadd_eval_cv_test = TRUE,
		stepadd_eval_aic_test = TRUE,
		stepadd_best_k = FALSE,
		stepdel_eval_cg_train = TRUE,
		stepdel_eval_igd_train = TRUE,
		stepdel_eval_aic_test = TRUE,
		stepdel_eval_cv_test = TRUE,
		stepdel_best_k = FALSE
	)
		
}

is.completed <- function(step.num) {

	set.id	<- GetOutputSetId(step.num)
	list.id <- GetOutputListId(step.num)
	
	if(set.id != MNULL) {
		set.val <- unlist(GetFeatureSet(set.id))
		if(is.na(set.val[1]) == TRUE) {
			return(FALSE)
		} else {
			return(TRUE)
		}
	} 
	
	if(list.id != MNULL) {
		list.val <- unlist(GetFeatureMapValues(list.id))
		if(is.na(list.val[1]) == TRUE) {
			return(FALSE)
		} else {
			return(TRUE)
		}
	
	}
	
}

GetFeatureIndices <- function(dataset.id) {
	valid.indices <- which(mdb$Features$dataset.id == dataset.id)
	valid.feat.indices <- as.integer(mdb$Features$feature.id[valid.indices])
	feat.indices <- setdiff(valid.feat.indices, mdb$Dataset[dataset.id]$target.index)
	return(feat.indices)
}

GetStepNumForFeatId <- function(feat.id) {
	return(mdb$FeatureSetDB[feat.id]$inst.id)
}

GetStepNumForListId <- function(list.id) {
	return(mdb$FeatureMapDB[list.id]$inst.id)
}

GetParents <- function(step.num) {
	
	parent.list <- c()

	feat.id <- GetInputSetId.1(step.num)
	if(feat.id != MNULL) {
		parent.step <- GetStepNumForFeatId(feat.id)
		parent.list <- union(parent.list, parent.step)
	}

	feat.id <- GetInputSetId.2(step.num)
	if(feat.id != MNULL) {
		parent.step <- GetStepNumForFeatId(feat.id)
		parent.list <- union(parent.list, parent.step)
	}

	list.id <- GetInputList.1(step.num) 
	if(list.id != MNULL) {
		parent.step <- GetStepNumForListId(list.id)
		parent.list <- union(parent.list, parent.step)
	}
	
	return(parent.list)
	
}		

GetChildren <- function(step.num) {
	
	child.list <- c()
	
	feat.id <- GetOutputSetId(step.num)
	if(feat.id != MNULL) {
		feat.id.occurences <- which(mdb$Program$inpset.id1 == feat.id)
		if(length(feat.id.occurences) > 0) 
			child.list <- union(child.list, feat.id.occurences)

		feat.id.occurences <- which(mdb$Program$inpset.id2 == feat.id)
		if(length(feat.id.occurences) > 0) 
			child.list <- union(child.list, feat.id.occurences)
	}
	
	list.id <- GetOutputListId(step.num)
	if(list.id != MNULL) {
		list.id.occurences <- which(mdb$Program$inplist.id1 == list.id)	
		if(length(list.id.occurences) > 0) 
			child.list <- union(child.list, list.id.occurences)
	}

	return(child.list)

}

GetStepsForDataset <- function(dataset.id) {
	steps.to.consider <- which(mdb$Program$dataset.id == dataset.id)
	return(steps.to.consider)
}

GenNodeList <- function(dataset.id) {
	
	# Iterate through the steps in the program table.
	# for each node, 
	# 1) Populate step.num
	# 2) Populate parents
	# 3) Populate children
	# 4) Populate level
	# 5) Populate data aware [ ?? ]
	# 6) Completed 

	node.list <- list()
	num.steps <- GetStepNum()
	steps.to.consider <- which(mdb$Program$dataset.id == dataset.id)
	
	if(length(steps.to.consider) == 0) {
		stop('There are no qualifying steps')
	}
	
	for(step.num in steps.to.consider) {
		node <- NodeClass(step.num)
		node[['data.aware']] <- is.data.aware(step.num)
		node[['parents']] <- GetParents(step.num)
		node[['children']] <- GetChildren(step.num)
		node[['level']] <- LEVEL_PENDING
		node[['completed']] <- is.completed(step.num)
		node[['step.num']] <- step.num
		node.list[[step.num]] <- node
	}
	return(node.list)
	
}

GetTargetIndex <- function(dataset.id) {
	
	target.index <- mdb$Dataset[dataset.id]$target.index
	return(target.index)
}

GetOrigNumFeatures <- function(dataset.id) {
	return(mdb$Dataset[dataset.id]$initial.num.features)
}

GetTotalNumFeatures <- function(dataset.id) {
	
	valid.indices <- which(mdb$Features$dataset.id == dataset.id)
	feat.indices <- mdb$Features$feature.id[valid.indices]
	target.index <- GetTargetIndex(dataset.id)
	feat.indices <- setdiff(valid.indices, target.index)
	tot.num.features <- max(as.integer(feat.indices)) + 1 # to account for 0 indexing in the UDF
	return(tot.num.features)
}

are.inputs.computed <- function(step.num) {
	
	is.feat.set.computed <- are.feat.inputs.computed(step.num)
	is.list.computed <- are.list.inputs.computed(step.num)
	return(is.feat.set.computed && is.list.computed)
}
are.feat.inputs.computed <- function(step.num) {
	feat.id <- GetInputSetId.1(step.num)
	if(feat.id != MNULL) {
		set.val <- GetFeatureSet(feat.id)
		if(is.na(set.val) == TRUE) 
			return(FALSE)
	} 
	
	feat.id <- GetInputSetId.2(step.num)
	if(feat.id != MNULL) {
		set.val <- GetFeatureSet(feat.id)
		if(is.na(set.val) == TRUE) 
			return(FALSE)
	} 
	return(TRUE)
	
}

are.list.inputs.computed <- function(step.num) {
	list.id <- GetInputList.1(step.num)
	if(list.id != MNULL) {
		list.val <- GetFeatureMapValues(list.id)
		if(is.na(list.val) == TRUE)
			return(FALSE)
	}
	return(TRUE)
}

GetOperatorId <- function(step.num) {
	return(mdb$Program[step.num]$op.id)
}

GetProgramCtr <- function(step.num) {
	return(mdb$Program[step.num]$program.ctr)
}

GetIgdConfig <- function(step.num, exit.flag=TRUE) {
	pgm.ctr <- GetProgramCtr(step.num)
	idx <- which(mdb$IgdConfig$step.num == step.num)
	if(length(idx) == 0) {
		 # Trying the program ctr
		idx <- which(mdb$IgdConfig$program.ctr == pgm.ctr)
		if(length(idx) == 0) {
			if(exit.flag == TRUE)
				stop('IgdConfig not found for the step.num/pgm.ctr')
			else
				return(MNULL)
		}
	}
	return(mdb$IgdConfig[idx])
}

GetCgConfig <- function(step.num, exit.flag = TRUE) {
	
	pgm.ctr <- GetProgramCtr(step.num)
	idx <- which(mdb$CgConfig$step.num == step.num)
	if(length(idx) == 0) {
		
		# Trying the program ctr
		idx <- which(mdb$CgConfig$program.ctr == pgm.ctr)
		if(length(idx) == 0)
			if(exit.flag == TRUE)
				stop('CgConfig not found for step.num/pgm.ctr')
			else
				return(MNULL)
	}
	return(mdb$CgConfig[idx])
}

GetBestKConfig <- function(step.num) {
	
	idx <- which(mdb$BestKConfig$step.num == step.num)
	if(length(idx) == 0) {
		stop('BestKConfig not found for step.num')
	}
	return(mdb$BestKConfig[idx])
}


SetUpForOperation <- function(step.num) {
	
	op.id <- GetOperatorId(step.num)
	if(op.id == "CORRELATION_X") 
		SetUpForCorrelationX(step.num)
	else if(op.id == "CORRELATION_Y")
		SetUpForCorrelationY(step.num)
	else if(op.id == "COEF_IGD_TRAIN")
		SetUpForIgdCoeffLearner(step.num)
	else if(op.id == "COEF_CG_TRAIN")
		SetUpForCgCoeffLearner(step.num)
	else if(op.id == "EVAL_CG_TRAIN")
		SetUpForEvalCgTrain(step.num)
	else if(op.id == "EVAL_IGD_TRAIN")
		SetUpForEvalIgdTrain(step.num)
	else if(op.id == "EVAL_CV_TEST")
		SetUpForEvalCvTest(step.num)
	else if(op.id == "EVAL_AIC_TEST")
		SetUpForEvalAiCTest(step.num)
	else if(op.id == "STEPADD_EVAL_IGD_TRAIN")
		SetUpForStepAddIgdTrain(step.num)
	else if(op.id == "STEPADD_EVAL_CG_TRAIN")
		SetUpForStepAddCgTrain(step.num)
	else if(op.id == "STEPADD_EVAL_CV_TEST")
		SetUpForStepAddCvTest(step.num)
	else if(op.id == "STEPADD_EVAL_AIC_TEST")
		SetUpForStepAddAiCTest(step.num)
	else if(op.id == "STEPDEL_EVAL_IGD_TRAIN")
		SetUpForStepDelIgdTrain(step.num)
	else if(op.id == "STEPDEL_EVAL_CG_TRAIN")
		SetUpForStepDelCgTrain(step.num)
	else if(op.id == "STEPDEL_EVAL_CV_TEST")
		SetUpForStepDelCvTest(step.num)
	else if(op.id == "STEPDEL_EVAL_AIC_TEST")
		SetUpForStepDelAiCTest(step.num)
	else
		stop('Operator does not have valid setup operation')
}

ExeAgnosticOperation <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	if(op.id == "ADD_FEATURE") 
		res <- ExeAddFeatureSet(step.num)
	else if(op.id == "DEL_FEATURE") 
		res <- ExeDelFeatureSet(step.num)
	else if(op.id == "INTERSECT_FEATURE")
		res <- ExeIntersectFeatureSet(step.num)
	else if(op.id == "CREATE_FEATURE")
		res <- ExeCreateFeature(step.num)
	else if(op.id == "BEST_K")
		res <- ExeBestK(step.num)
	else if(op.id == "STEPADD_BEST_K")
		res <- ExeStepAddBestK(step.num)
	else if(op.id == "STEPDEL_BEST_K")
		res <- ExeStepDelBestK(step.num)
	else
		stop('Operator does not have a valid data agnostic operation')
		
			
}

GetExeObject <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	if(op.id == "CORRELATION_X") 
		obj <- bCreateObj("descstat.corr")
	else if(op.id == "CORRELATION_Y")
		obj <- bCreateObj("descstat.corr")
	else if(op.id == "COEF_IGD_TRAIN")
		obj <- bCreateObj("learner.lr")
	else if(op.id == "COEF_CG_TRAIN")
		obj <- bCreateObj("learner.cg")
	else if(op.id == "EVAL_CG_TRAIN")
		obj <- bCreateObj("learner.cg")
	else if(op.id == "EVAL_IGD_TRAIN")
		obj <- bCreateObj("learner.lr")
	else if(op.id == "EVAL_CV_TEST")
		obj <- bCreateObj("scorer.lr") # TODO(pradap) : Change to scorer.cv
	else if(op.id == "EVAL_AIC_TEST")
		obj <- bCreateObj("scorer.aic") # TODO(pradap) : Change to scorer.cv
	else if(op.id == "STEPADD_EVAL_IGD_TRAIN")
		obj <- bCreateObj("learner.lr")
	else if(op.id == "STEPADD_EVAL_CG_TRAIN")
		obj <- bCreateObj("learner.cg")
	else if(op.id == "STEPADD_EVAL_CV_TEST")
		obj <- bCreateObj("scorer.lr")
	else if(op.id == "STEPADD_EVAL_AIC_TEST")
		obj <- bCreateObj("scorer.aic")
	else if(op.id == "STEPDEL_EVAL_IGD_TRAIN")
		obj <- bCreateObj("learner.lr")
	else if(op.id == "STEPDEL_EVAL_CG_TRAIN")
		obj <- bCreateObj("learner.cg")
	else if(op.id == "STEPDEL_EVAL_CV_TEST")
		obj <- bCreateObj("scorer.lr")
	else if(op.id == "STEPDEL_EVAL_AIC_TEST")
		obj <- bCreateObj("scorer.aic")
	else
		stop('Operator does not have valid object create operation')
}

GetNewFeatString <- function(dataset.id) {
	valid.indices <- which(mdb$DerivedFeatures$dataset.id == dataset.id)
	s <- ""
	cnt <- 1

	for(i in valid.indices) {
		new.feat.id <- mdb$DerivedFeatures[i]$newfeature.id
		feat.id1 <- mdb$DerivedFeatures[i]$feature.id1
		feat.id2 <- mdb$DerivedFeatures[i]$feature.id2
		
		s <- paste(s, new.feat.id, "~", feat.id1, ",", feat.id2, sep="")
		if(cnt != length(valid.indices))
			s <- paste(s, "#", sep="")
		cnt <- cnt + 1
	}
	return(s)
}

GetNewFeatures <- function(dataset.id) {
	valid.indices <- which(mdb$DerivedFeatures$dataset.id == dataset.id)
	return(mdb$DerivedFeatures[valid.indices]$newfeature.id)
	
}
GetRelationName <- function(decode.indices, dataset.id, proposed.indices) {

	
	new.features <- GetNewFeatures(dataset.id)
	decode.indices <- setdiff(decode.indices, new.features)
	proposed.indices <- setdiff(proposed.indices, new.features)
	
	# Check in the derived datasets table.
	table.name <- GetExactMatchRelation(decode.indices, dataset.id)
	
	if(is.null(table.name) == FALSE) {
		return(table.name)
	}
	
	table.name <- GetThinnerRelation(decode.indices, dataset.id, proposed.indices) 
	
	if(is.null(table.name) == FALSE) {
		return(table.name)
	}

	table.name <- GetExactMatchRelation(proposed.indices, dataset.id)
	
	if(is.null(table.name) == FALSE) {
		return(table.name)
	}

	# Do Materialization
	table.name <- GetMaterializedTableName(proposed.indices, dataset.id)
	
	return(table.name)
}


GetExactMatchRelation <- function(decode.indices, dataset.id) {
	
	valid.indices <- which(mdb$DerivedDatasets$dataset.id == dataset.id)
	table.name <- NULL
	
	for(i in valid.indices) {
		feat.indices <- unlist(mdb$DerivedDatasets[i]$feature.indices)
		if(length(feat.indices) == length(decode.indices)) {
			if(all(sort(feat.indices) == sort(decode.indices)) == TRUE) {
				table.name <- mdb$DerivedDatasets[i]$table.name
			}
		}
	}
	return(table.name)
}

GetThinnerRelation <- function(decode.indices, dataset.id, proposed.indices) {
	
	valid.indices <- which(mdb$DerivedDatasets$dataset.id == dataset.id)
	table.name <- NULL
	cur.length <- length(proposed.indices) + 1
	
	for(i in valid.indices) {
		feat.indices <- unlist(mdb$DerivedDatasets[i]$feature.indices)
		len.feat.indices <- length(feat.indices)
		
		if(len.feat.indices > length(decode.indices) && len.feat.indices <=
								length(proposed.indices)) {
			int.set <- intersect(decode.indices, feat.indices)
			if(length(int.set) == length(decode.indices)) {
				if(all(sort(int.set) == sort(decode.indices)) == TRUE) {
					if(len.feat.indices < cur.length) {
						table.name <- mdb$DerivedDatasets[i]$table.name
						cur.length <- len.feat.indices
					}
				}
			}
			
		}
		
	}
	return(table.name)
	
}



GetMaterializedTableName <- function(proposed.indices, dataset.id) {
	inp.table.name <- mdb$Dataset[dataset.id]$name
	target.index <- GetTargetIndex(dataset.id)
	
	num.datasets <- nrow(mdb$DerivedDatasets[dataset.id])
	
	out.table.name <- paste(inp.table.name, num.datasets, sep="_")
	feature.names <- GetFeatureNames(c(proposed.indices, target.index), dataset.id)
	feature.names <- as.character(paste(feature.names, collapse=","))	
	
	# Update derived datasets table.
	InsertRowInTable('DerivedDatasets', data.table(dataset.id, list(proposed.indices), out.table.name))
	
	# DL code start
	bMaterialize(inp.table.name, feature.names, out.table.name)
	# DL code end

	
	return(out.table.name)

	
}

ConsumeResult <- function(step.num, res.string) {
	
	op.id <- GetOperatorId(step.num)
	if(op.id == "CORRELATION_X")
		ResCorrelationX(step.num, res.string)
	else if(op.id == "CORRELATION_Y")
		ResCorrelationY(step.num, res.string)
	else if(op.id == "COEF_IGD_TRAIN")
		ResIgdCoeffLearner(step.num, res.string)
	else if(op.id == "COEF_CG_TRAIN")
		ResCgCoeffLearner(step.num, res.string)
	else if(op.id == "EVAL_IGD_TRAIN")
		ResEvaluateIgdTrain(step.num, res.string)
	else if(op.id == "EVAL_CG_TRAIN")
		ResEvaluateCgTrain(step.num, res.string)
	else if(op.id == "EVAL_CV_TEST")
		ResEvaluateCvTest(step.num, res.string)	
	else if(op.id == "EVAL_AIC_TEST")
		ResEvaluateAiCTest(step.num, res.string)	
	else if(op.id == "STEPADD_EVAL_IGD_TRAIN")
		ResStepAddIgdTrain(step.num, res.string)
	else if(op.id == "STEPADD_EVAL_CG_TRAIN")
		ResStepAddCgTrain(step.num, res.string)
	else if(op.id == "STEPADD_EVAL_CV_TEST")
		ResStepAddCvTest(step.num, res.string)
	else if(op.id == "STEPADD_EVAL_AIC_TEST")
		ResStepAddAiCTest(step.num, res.string)
	else if(op.id == "STEPDEL_EVAL_IGD_TRAIN")
		ResStepDelIgdTrain(step.num, res.string)
	else if(op.id == "STEPDEL_EVAL_CG_TRAIN")
		ResStepDelCgTrain(step.num, res.string)
	else if(op.id == "STEPDEL_EVAL_CV_TEST")
		ResStepDelCvTest(step.num, res.string)
	else if(op.id == "STEPDEL_EVAL_AIC_TEST")
		ResStepDelAiCTest(step.num, res.string)
	else
		stop('There is no operation to consume the result')
}

GetDatasetIdFromStepNum <- function(step.num) {
	return(mdb$Program[step.num]$dataset.id)
}

UpdateLevelForNode <- function(step.num, node.list) {
	
	# The level for a node is updated as follows:
	# initial.condition : LEVEL_PENDING
	#
	# If the operation corresponding to the node is not computed
	# then, get the max of the levels of the parents and add 1 
	# to it.

	node <- node.list[[step.num]]
	if(node$completed == 1) {
		mx.level <- 0	
	}
	
	if(are.feat.inputs.computed(step.num) == FALSE) {
		return(LEVEL_PENDING)
	}

	# update the level based on its parents.
	parents <- node$parents
	mx.level <-  0
	
	if(length(parents) == 0) {
		mx.level <- 0
	} else {	
		for(p.id in parents) {
			parent.node <- node.list[[p.id]]
			if(node$completed == 0) {
				level <- parent.node$level
				if(level != LEVEL_PENDING) {
					mx.level <- ifelse(max(level + 1, mx.level) > mx.level, level + 1, mx.level)
				
				} else {
					return(level)
				}
			}
		}
	}
	return(mx.level)
}

UpdateLevels <- function(node.list, dataset.id) {
	steps.to.consider <- which(mdb$Program$dataset.id == dataset.id)
	
	if(length(steps.to.consider) == 0) {
		stop('There are no qualifying steps')
	}

	for(step.num in steps.to.consider) {
		level <- UpdateLevelForNode(step.num, node.list)
		node.list[[step.num]]$level <- level
	}

	return(node.list)
}		


is.matopt <- function() {
	return(TRUE)
}
is.stepadd.op <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	b <- grep("STEPADD", op.id, ignore.case = TRUE)
	return(isTRUE(b == 1))
}

GetDecodeIndicesMap <- function(table.name, dataset.id, decode.indices) {
	
	new.features <- GetNewFeatures(dataset.id)
	decode.indices <- setdiff(decode.indices, new.features)

	valid.derived.datasets <- which(mdb$DerivedDatasets$dataset.id == dataset.id)
	if(length(valid.derived.datasets) == 0) {
		stop('No valid derived datasets for the given dataset id')
	}	
	
	derived.datasets.table.names <- tolower(mdb$DerivedDatasets[valid.derived.datasets]$table.name)
	valid.idx <- which(derived.datasets.table.names ==  tolower(table.name)) 
	if(length(valid.idx) != 1) {
		stop('Input table.name matches none/more than one table.name from the derived
datasets')
	}
	valid.idx <- valid.derived.datasets[valid.idx]
	feature.indices <- unlist(mdb$DerivedDatasets[valid.idx]$feature.indices)
	map.feat.indices <- match(decode.indices, feature.indices)
	return(map.feat.indices - 1) # to account for indexing from 0
}


GetTargetIndexFromDerivedDataset <- function(table.name, dataset.id) {
	valid.derived.datasets <- which(mdb$DerivedDatasets$dataset.id == dataset.id)
	if(length(valid.derived.datasets) == 0) {
		stop('No valid derived datasets for the given dataset id')
	}	
	
	derived.datasets.table.names <- tolower(mdb$DerivedDatasets[valid.derived.datasets]$table.name)
	valid.idx <- which(derived.datasets.table.names ==  tolower(table.name)) 
	valid.idx <- valid.derived.datasets[valid.idx]
	if(length(valid.idx) != 1) {
		stop('Input table.name matches none/more than one table.name from the derived datasets')
	}
	feature.indices <- unlist(mdb$DerivedDatasets[valid.idx]$feature.indices)
	return(length(feature.indices))
	
	
}



GetReadCostForAllFeatures <- function(dataset.id) {
	return(31*4)
}
GetWriteCostForAllFeatures <- function(dataset.id) {
	return(31*8)
}
GetReadIntercept <- function(dataset.id) {
	return(5)
}

GetWriteIntercept <- function(dataset.id) {
	return(500)
}

GetProgramId <- function() {
	#return(2)
	return(mdb$program.id)
	
}

SetProgramId <- function(program.id) {
	mdb$program.id <- program.id
}


GetSessionId <- function() {
	#return(2)
	return(mdb$session.id)
}

SetSessionId <- function(session.id) {
	mdb$session.id <- session.id
}


# ---------------- stat --------------
is.descstat.op <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	if(op.id == "CORRELATION_X")
		return(TRUE)
	else if(op.id == "CORRELATION_Y")
		return(TRUE)
	else if(op.id == "COEF_CG_TRAIN")
		return(TRUE)
	else if(op.id == "COEF_IGD_TRAIN")
		return(TRUE)
	else 
		return(FALSE)
}

is.eval.op <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	if(op.id == "EVAL_CG_TRAIN")
		return(TRUE)
	else if(op.id == "EVAL_IGD_TRAIN")
		return(TRUE)
	else if(op.id == "EVAL_CV_TEST")
		return(TRUE)
	else if(op.id == "EVAL_AIC_TEST")
		return(TRUE)
	else
		return(FALSE)
}
is.expl.op <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	if(op.id == "STEPADD_EVAL_IGD_TRAIN")
		return(TRUE)
	else if(op.id == "STEPADD_EVAL_CG_TRAIN")
		return(TRUE)
	else if(op.id == "STEPADD_EVAL_CV_TEST")
		return(TRUE)
	else if(op.id == "STEPADD_EVAL_AIC_TEST")
		return(TRUE)
	else if(op.id == "STEPDEL_EVAL_IGD_TRAIN")
		return(TRUE)
	else if(op.id == "STEPDEL_EVAL_CG_TRAIN")
		return(TRUE)
	else if(op.id == "STEPDEL_EVAL_CV_TEST")
		return(TRUE)
	else if(op.id == "STEPDEL_EVAL_AIC_TEST")
		return(TRUE)
	else
		return(FALSE)
}

# ------------------ prov --------------


ProvDelegator <- function(step.num) {
	op.id <- GetOperatorId(step.num)
	print(op.id)
	if(op.id == "CORRELATION_X") 
		StoreCorrelationXResult(step.num)
	else if(op.id == "CORRELATION_Y")
		StoreCorrelationYResult(step.num)	
	else if(op.id == "COEF_IGD_TRAIN")
		StoreIgdCoeff(step.num)
	else if(op.id == "COEF_CG_TRAIN")
		StoreCgCoeff(step.num)
	else if(op.id == "EVAL_CG_TRAIN")
		StoreEvalCgCoeff(step.num)
	else if(op.id == "EVAL_IGD_TRAIN")
		StoreEvalIgdCoeff(step.num)
	else if(op.id == "EVAL_CV_TEST")
		StoreEvalCv(step.num)
	else if(op.id == "EVAL_AIC_TEST")
		StoreEvalAiC(step.num)
	else if(op.id == "STEPADD_EVAL_IGD_TRAIN")
		StoreStepAddEvalIgd(step.num)
	else if(op.id == "STEPADD_EVAL_CG_TRAIN")
		StoreStepAddEvalCg(step.num)
	else if(op.id == "STEPADD_EVAL_CV_TEST")
		StoreStepAddEvalCv(step.num)
	else if(op.id == "STEPADD_EVAL_AIC_TEST")
		StoreStepAddEvalAiC(step.num)
	else if(op.id == "STEPDEL_EVAL_IGD_TRAIN")
		StoreStepDelEvalIgd(step.num)
	else if(op.id == "STEPDEL_EVAL_CG_TRAIN")
		StoreStepDelEvalCg(step.num)
	else if(op.id == "STEPDEL_EVAL_CV_TEST")
		StoreStepdelEvalCv(step.num)
	else if(op.id == "STEPDEL_EVAL_AIC_TEST")
		StoreStepdelEvalAiC(step.num)
	else
		stop('Operator does not have valid store operation')
}

StoreProvResults <- function() {
	steps.to.consider <- 1 : nrow(mdb$Program)
	for(step.num in steps.to.consider) {
		if(is.data.aware(step.num)) {
			res <- ProvDelegator(step.num)
		}	
	}	
}

GetMatchingOps <- function(step.num, dataset.id) {
	
	out.list <- c()
	op.id <- GetOperatorId(step.num)
	match.indices <- which(mdb$Program$op.id == op.id)
	valid.indices <- which(match.indices < step.num)


	for(index in match.indices[valid.indices]) {
		if((is.data.aware(step.num) == TRUE) && (mdb$Program[step.num]$dataset.id == mdb$Program[index]$dataset.id)) {
			list.id <- GetOutputListId(index)
			if(list.id != MNULL) {
				values <- GetFeatureMapValues(list.id)
				is.computed <- !all(is.na(unlist(values)))
				if(is.computed) {
					out.list <- c(out.list, index)
				}
			} else {
				set.id <- GetOutputSetId(step.num)
				if(set.id != MNULL) {
					set.val <- GetFeatureSet(set.id)
					is.computed <- !all(is.na(unlist(values)))
					if(is.computed) {
						out.list <- c(out.list, index)
					}
				}
			}
			
		}
		
	}
	return(out.list)
	
}

are.featsets.same <- function(feat.id1, feat.id2) {
	if(feat.id1 == feat.id2)
		return(TRUE)
	# If one of the feat ids is MNULL then it is not a match.
	if(feat.id1 == MNULL || feat.id2 == MNULL) 
		return(FALSE)
	set.1 <- unlist(GetFeatureSet(feat.id1))
	set.2 <- unlist(GetFeatureSet(feat.id2))
	
	if(is.na(set.1) || is.na(set.2))
		return(FALSE)
	
	return(VectorMatch(sort(set.1), sort(set.2)))
	
}

are.featlists.same <- function(list.id1, list.id2) {
	if(list.id1 == list.id2)
		return(TRUE)
	
	if(list.id1 == MNULL || list.id2 == MNULL)
		return(FALSE)
	
	list.1 <- as.numeric(unlist(GetFeatureMapValues(list.id1)))
	list.2 <- as.numeric(unlist(GetFeatureMapValues(list.id2)))
	
	feat.list.1 <- unlist(GetFeatureMapFeatureSet(list.id1))
	feat.list.2 <- unlist(GetFeatureMapFeatureSet(list.id2))
	
	bool.res <- VectorMatch(feat.list.1, feat.list.2)
	if(bool.res == FALSE)
		return(FALSE)

	bool.res <- ApproxVectorMatch(list.1, list.2)
	return(bool.res)

}

GetParamConfig <- function(step.num) {
	config <- GetIgdConfig(step.num, exit.flag = FALSE)
	is.cg <- FALSE
	if((is.list(config) == FALSE) && (config == MNULL)) {
		config <- GetCgConfig(step.num, exit.flag = FALSE)
		is.cg <- TRUE
	}
	return(list(config = config, is.cg = is.cg))
}

are.vals.same <- function(val.1, val.2) {
	if(val.1 != val.2)
		return(FALSE)
	else
		return(TRUE)
}
are.igd.config.same <- function(config.1, config.2) {
	
	bool.res <- are.vals.same(config.1$dataset.id, config.2$dataset.id)
	if(bool.res == FALSE)
		return(FALSE)
	
#	bool.res <- are.vals.same(config.1$num.iters, config.2$num.iters)
#	if(bool.res == FALSE)
#		return(FALSE)
	
	bool.res <- are.vals.same(config.1$step.size, config.2$step.size)
	if(bool.res == FALSE)
		return(FALSE)

	
	bool.res <- are.vals.same(config.1$decay, config.2$decay)
	if(bool.res == FALSE)
		return(FALSE)

	bool.res <- are.vals.same(config.1$init.wt, config.2$init.wt)
	if(bool.res == FALSE)
		return(FALSE)


	bool.res <- are.vals.same(config.1$num.folds, config.2$num.folds)
	if(bool.res == FALSE)
		return(FALSE)

	return(TRUE)
		
	
}

are.cg.config.same <- function(config.1, config.2) {

	bool.res <- are.vals.same(config.1$dataset.id, config.2$dataset.id)
	if(bool.res == FALSE)
		return(FALSE)
	
	#bool.res <- are.vals.same(config.1$num.iters, config.2$num.iters)
	#if(bool.res == FALSE)
	#	return(FALSE)

	bool.res <- are.vals.same(config.1$init.wt, config.2$init.wt)
	if(bool.res == FALSE)
		return(FALSE)


	bool.res <- are.vals.same(config.1$num.folds, config.2$num.folds)
	if(bool.res == FALSE)
		return(FALSE)

	return(TRUE)
}


are.params.same <- function(step.num.1, step.num.2) {
	param.1 <- GetParamConfig(step.num.1)
	param.2 <- GetParamConfig(step.num.2)
	
	config.1 <- param.1$config
	is.cg.1 <- param.1$is.cg

	config.2 <- param.2$config
	is.cg.2 <- param.2$is.cg

	if(length(config.1) != length(config.2)) 
		return(FALSE)
	
	if(is.cg.1 != is.cg.2)
		return(FALSE)

	if((config.1 == MNULL) && (config.2 == MNULL))
		return(TRUE)
	
	if(is.cg.1 == TRUE)
		bool.res <- are.cg.config.same(config.1, config.2)

	else
		bool.res <- are.igd.config.same(config.1, config.2)

	return(bool.res)
	
}

GetReusableStep <- function(step.num.list, inp.step.num) {
	inp.set.id1 <- GetInputSetId.1(inp.step.num)
	inp.set.id2 <- GetInputSetId.2(inp.step.num)
	inp.list.id <- GetInputList.1(inp.step.num)	

	
	for(step.num in step.num.list) {
			
		set.id1 <- GetInputSetId.1(step.num)
		set.id2 <- GetInputSetId.2(step.num)
		list.id <- GetInputList.1(step.num)
		bool <- c()

		bool[1] <- are.featsets.same(inp.set.id1, set.id1)
		bool[2] <- are.featsets.same(inp.set.id2, set.id2)
		bool[3] <- are.featlists.same(inp.list.id, list.id)
		bool[4] <- are.params.same(inp.step.num, step.num) 


		if(all(bool) == TRUE)
			return(step.num)
	}
	return(MNULL)
}
CheckAndReuseLocalResultForAStep <- function(step.num, dataset.id) {
	if(are.inputs.computed(step.num)) {
		matching.step.nums <- GetMatchingOps(step.num, dataset.id)
		if(length(matching.step.nums) == 0)
			return(FALSE)
		reusable.step <- GetReusableStep(matching.step.nums, step.num)
		if(reusable.step == MNULL)
			return(FALSE)
		else {
			# There is a match
			from.set.id <- GetOutputSetId(reusable.step)
			if(from.set.id != MNULL) {
				to.set.id <- GetOutputSetId(step.num)
				CopyFeatureSet(from.set.id, to.set.id)
			} else {
				from.list.id <- GetOutputListId(reusable.step)
				if(from.list.id != MNULL){
					to.list.id <- GetOutputListId(step.num)
					CopyFeatureMapFeatSet(from.list.id, to.list.id)
					CopyFeatureMapValues(from.list.id, to.list.id)
				}
			}
		
			return(TRUE)
		}
		
	}
	return(FALSE)
}
       
 
CheckAndReuseLocalResults <- function(node.list, dataset.id) {
	steps.to.consider <- which(mdb$Program$dataset.id == dataset.id)
	are.results.updated <- FALSE
	for(step.num in steps.to.consider) {
		node <- node.list[[step.num]]
		if(node$completed == 0 && node$data.aware == 1) {
			res <- CheckAndReuseLocalResultForAStep(step.num, dataset.id)
			if(res == TRUE) {
				node.list[[step.num]]$completed = 1
				are.results.updated <- TRUE
			}
		}
	
	}
	return(list(are.results.updated = are.results.updated, node.list = node.list))
}

GetProgramname <- function() {
	return(mdb$program.name)
}

SetProgramname <- function(program.name) {
	mdb$program.name <- program.name
}

GetNumFeatureMapEntries <- function() {
	return(nrow(mdb$FeatureMapDB))
}

GetDataset <- function(dataset.id) {
	if(dataset.id > nrow(mdb$Dataset) || dataset.id <= 0)
		stop('Argument dataset.id is invalid')
	return(mdb$Dataset[dataset.id])
}


GetDatasetnames <- function() {
	return(mdb$Dataset$name)
}
GetMinMaxForAFeature <- function(feat.name, dataset.id) {
	valid.dataset.indices <- which(mdb$Features$dataset.id == dataset.id)
	idx <- which(tolower(mdb$Features[valid.dataset.indices]$feature.name) == tolower(feat.name))
	 reqd.idx <- valid.dataset.indices[idx]
	  
	 min.val <-  mdb$Features[reqd.idx]$min.val
	 max.val <-  mdb$Features[reqd.idx]$max.val
	 return(list(min.val = min.val, max.val = max.val))
			
}

is.updateprogram <- function() {
	return(TRUE)
}


AddProgramLine <- function(pgm.str) {
	num.lines <- length(mdb$program.lines)
	new.line <- num.lines + 1
	mdb$program.lines[[new.line]] <- pgm.str
	
}

ExecuteProgram <- function(program.name, dataset.id) {
		SetInteractive(FALSE)
		pgm.text <- GetPgmTextFromName(program.name)
		print("Executing in Batchmode")
		new.pattern <- paste0('dataset.id =', dataset.id)
		tt <- grep('colpgm', program.name)
		pgm.text <- sub('dataset.id = !!', new.pattern, pgm.text)
		writeLines(pgm.text, con='.temp.R')
		program.name <- '.temp.R'
		time.taken <- system.time(BatchMode(program.name), gcFirst = TRUE)
		printf("Batchmode took %f seconds", time.taken[[3]])
		print(paste("Batch mode execution took ", time.taken[[3]], "seconds", sep=""))
		SetInteractive(TRUE)
}
ViewProgram <- function(program.name, dataset.id = 1) {
	pgm.txt <- GetPgmTextFromName(program.name)
	new.pattern <- paste0('dataset.id =', dataset.id)
	pgm.txt <- sub('dataset.id = !!', new.pattern, pgm.txt)
	print(pgm.txt)

}
