#Copyright 2013 Pradap  and Arun Kumar and Christopher Re
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
#
#
# FeatureSetClass: Class definition to represent a feature set.
# Constructor Args:
#  id: Unique id for a feature set. Equivalently, it is the "step num" in the program table.
#  inp.set: Input set value. It is represented as an integer vector. Integers uniquely map to features in the dataset.
FeatureSetClass <- function(id, inp.set) {
	
	# Create a list of input attributes
	nc <- list(feat.id = id, set.val = inp.set)
	nc <- list2env(nc)
	
	# Set the classname
	class(nc) <- "FeatureSetClass"
	return(nc)
}

# FeatureMapClass: Class definition to represent a feature map. A feature map, maps a set of feature sets to numeric values.
# Constructor Args:
#  id: Unique id for a feature set. Equivalently, it is the "step num" in the program table.
#  feature.list: List of feature sets. 
#  values: List of values corresponding to the feature sets.

FeatureMapClass <- function(id, feature.list, values) {
	
	# Create a list of input attributes
	nc <- list(list.id = id, feature.list = feature.list, values = values)
	nc <- list2env(nc)

	# Set the classname
	class(nc) <- "FeatureMapClass"
	return(nc)
} 


# NodeClass: Class definition to represent a node in the program DAG. Each node is an operation defined in the columbus framework.
# Constructor Args:
#	step.num : Step number in the program table.

NodeClass <- function(step.num) {

	# Create a list of attributes for the node class.
	nc <- list(step.num = step.num, # Unique id for the node.
		   level = NA,    	# Denotes the level in the program DAG. The edges in the DAG are formed by programmatic dependencies.
		   parents = c(), 	# Vector of parent nodes for the current node. 
		   children = c(),      # Vector of childrent nodes for the current node.
		   data.aware = FALSE,  # Boolean indicating whether the current node is data aware or not.
		   completed = FALSE    # Boolean indicating whether the current node has completed its execution.
	      )
	nc <- list2env(nc)

	# Set the classname
	class(nc) <- "NodeClass"
	return(nc)
}


# Print functions for the Node class
# Args:
#  Object of type NodeClass

print.NodeClass <- function(x) {
	
	# Error handling
	if(class(x) != "NodeClass") {
		stop("Argument is not of type NodeClass")
	}
	printf("step.num   : %d", x[["step.num"]])
	printf("level      : %s", x[["level"]])
	printf("data.aware : %d", x[["data.aware"]])
	printf("parents    : %s", list(x[["parents"]]))
	printf("children   : %s", list(x[["children"]]))
	printf("completed  : %d", x[["completed"]])
}

# Print function for the FeatureMap
# Args:
#  Object of type FeatureMap

print.FeatureMapClass <- function(x) {
	
	# Error handling
	if(class(x) != "FeatureMapClass") {
		stop("Argument is not of type FeatureMapClass")
	}
	list.id <- x$list.id
	feature.list <- x$feature.list
	
    	cat(sprintf('\n'))
    	cat(paste("FeatureList Id : ", list.id))
    	cat(sprintf('\n'))
    	cat(paste("FeatureList : ", feature.list))
    	cat(sprintf('\n'))
	step.num <- GetStepNumForListId(list.id)
	dataset.id <- GetDatasetIdFromStepNum(step.num)
	
    	cat(paste("FeatureSet Vals : ", paste(GetFeatureNames(unlist(feature.list), dataset.id), collapse=",")))
    	cat(sprintf('\n'))
    #cat(paste("Values : ", list(unlist(x$values))))
    	cat(paste("Values : ", list(unlist(mdb$FeatureMapDB[list.id]$values))))
    	cat(sprintf('\n'))
}


# Print function for the Featureset
# Args:
#  Object of type FeatureSet class

print.FeatureSetClass <- function(x) {
	
	# Error handling
	if(class(x) != "FeatureSetClass") {
		stop("Argument is not of type FeatureSetClass")
	}
	step.num <- GetStepNumForFeatId(x$feat.id)
	dataset.id <- GetDatasetIdFromStepNum(step.num)
	
	printf("FeatureSet Id : %d", x[["feat.id"]])
	printf("Set values    : %s", list(as.numeric(unlist(x[["set.val"]])))) # Replace with feature names.
	feature.names <- GetFeatureNames(as.numeric(unlist(x[["set.val"]])), dataset.id)
	printf("Feature names : %s", paste(feature.names, collapse=","))
}
